from databroker.assets.handlers import HandlerBase
import h5py

class HXNHandlerBase(HandlerBase):

    def __init__(self, resource_fn, **kwargs):
        if resource_fn.startswith('/data'):
            resource_fn = '/nsls2/data/hxn/legacy' + resource_fn[5:]
        self._handle = h5py.File(resource_fn, "r", libver='latest', swmr=True, locking=False)

    @property
    def filename(self):
        return self._handle.filename




from __future__ import print_function
import logging
from pathlib import PurePath
from ophyd import (AreaDetector, CamBase, TIFFPlugin, Component as Cpt,
                   HDF5Plugin, Device, StatsPlugin, ProcessPlugin,
                   ROIPlugin, TransformPlugin, EpicsSignal)
from ophyd.areadetector import EpicsSignalWithRBV

from ophyd.areadetector.base import ADComponent
from ophyd.areadetector.filestore_mixins import (
    FileStoreTIFF, FileStorePluginBase)

from .utils import makedirs
from .trigger_mixins import (HxnModalTrigger, FileStoreBulkReadable)


logger = logging.getLogger(__name__)


class MerlinTiffPlugin(TIFFPlugin, FileStoreBulkReadable, FileStoreTIFF,
                       Device):
    def mode_external(self):
        total_points = self.parent.mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

    def get_frames_per_point(self):
        mode = self.parent.mode_settings.mode.get()
        if mode == 'external':
            return 1
        else:
            return self.parent.cam.num_images.get()


class MerlinDetectorCam(CamBase):
    acquire = ADComponent(EpicsSignal,'Acquire')
    quad_merlin_mode = ADComponent(EpicsSignalWithRBV,'QuadMerlinMode')
    pass


class MerlinDetector(AreaDetector):
    cam = Cpt(MerlinDetectorCam, 'cam1:',
              read_attrs=[],
              configuration_attrs=['image_mode', 'trigger_mode',
                                   'acquire_time', 'acquire_period'],
              )


class MerlinFileStoreHDF5(FileStorePluginBase, FileStoreBulkReadable):
    _spec = 'TPX_HDF5'
    filestore_spec = _spec

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs.update([(self.file_template, '%s%s_%6.6d.h5'),
                                (self.file_write_mode, 'Stream'),
                                (self.compression, 'zlib'),
                                (self.capture, 1)
                                ])

    def stage(self):
        logger.info("Staging")
        staged = super().stage()
        logger.info("Staging step 2")
        res_kwargs = {'frame_per_point': 1}
        logger.info("res_kwargs = {frame_per_point: }")
        
        logger.debug("Inserting resource with filename %s", self._fn)
        logger.info("Inserting resource with filename %s", self._fn)
        self._generate_resource(res_kwargs)
        logger.info("generating resources")
        logger.info("Staged")
        return staged

    def make_filename(self):
        fn, read_path, write_path = super().make_filename()
        mode_settings = self.parent.mode_settings
        if mode_settings.make_directories.get():
            makedirs(read_path)
        return fn, read_path, write_path


class HDF5PluginWithFileStore(HDF5Plugin, MerlinFileStoreHDF5):
    def stage(self):
        mode_settings = self.parent.mode_settings
        total_points = mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

        # ensure that setting capture is the last thing that's done
        self.stage_sigs.move_to_end(self.capture)
        return super().stage()


class HxnMerlinDetector(HxnModalTrigger, MerlinDetector):
    hdf5 = Cpt(HDF5PluginWithFileStore, 'HDF1:',
               read_attrs=[],
               configuration_attrs=[],
               write_path_template='/data/%Y/%m/%d/',
               root='/data')

    proc1 = Cpt(ProcessPlugin, 'Proc1:')
    stats1 = Cpt(StatsPlugin, 'Stats1:')
    stats2 = Cpt(StatsPlugin, 'Stats2:')
    stats3 = Cpt(StatsPlugin, 'Stats3:')
    stats4 = Cpt(StatsPlugin, 'Stats4:')
    stats5 = Cpt(StatsPlugin, 'Stats5:')
    transform1 = Cpt(TransformPlugin, 'Trans1:')
    roi1 = Cpt(ROIPlugin, 'ROI1:')
    roi2 = Cpt(ROIPlugin, 'ROI2:')
    roi3 = Cpt(ROIPlugin, 'ROI3:')
    roi4 = Cpt(ROIPlugin, 'ROI4:')

    # tiff1 = Cpt(MerlinTiffPlugin, 'TIFF1:',
    #             read_attrs=[],
    #             configuration_attrs=[],
    #             write_path_template='/data/%Y/%m/%d/',
    #             root='/data')

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 **kwargs):
        if read_attrs is None:
            read_attrs = ['hdf5', 'cam']
        if configuration_attrs is None:
            configuration_attrs = ['hdf5', 'cam']

        if 'hdf5' not in read_attrs:
            # ensure that hdf5 is still added, or data acquisition will fail
            read_attrs = list(read_attrs) + ['hdf5']

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, **kwargs)

    def mode_internal(self):
        super().mode_internal()

        count_time = self.count_time.get()
        if count_time is not None:
            self.stage_sigs[self.cam.acquire_time] = count_time
            self.stage_sigs[self.cam.acquire_period] = count_time + 0.005

    def mode_external(self):
        super().mode_external()

        # NOTE: these values specify a debounce time for external triggering so
        #       they should be set to < 0.5 the expected exposure time, or at
        #       minimum the lowest possible dead time = 1.64ms
        expected_exposure = 0.001
        min_dead_time = 0.00164
        self.stage_sigs[self.cam.acquire_time] = expected_exposure
        self.stage_sigs[self.cam.acquire_period] = expected_exposure + min_dead_time

        self.cam.stage_sigs[self.cam.trigger_mode] = 'Trigger Enable'



from .hxn_handler import HXNHandlerBase
from databroker.assets.handlers import HandlerBase
import h5py

class SISHDF5Handler(HXNHandlerBase):
    HANDLER_NAME = "SIS_HDF51_FLY_STREAM_V1"

    def __init__(self, resource_fn, *, frame_per_point):
        self._frame_per_point = frame_per_point
        super().__init__(resource_fn)

    def __call__(self, *, column, point_number):
        n_first = point_number * self._frame_per_point
        n_last = n_first + self._frame_per_point
        ds = self._handle[column]
        ds.id.refresh()
        return ds[n_first:n_last]

    # def close(self):
    #     self._handle.close()
    #     self._handle = None
    #     super().close()
class BulkMerlinStream(HXNHandlerBase):
    HANDLER_NAME = 'MERLIN_FLY_STREAM_V2'

    def __init__(self, resource_fn, *, frame_per_point):
        self._frame_per_point = frame_per_point
        super().__init__(resource_fn)

    def __call__(self, point_number):
        n_first = point_number * self._frame_per_point
        n_last = n_first + self._frame_per_point
        ds = self._handle['entry/instrument/detector/data']
        ds.id.refresh()
        return ds[n_first:n_last, :, :]

    def dataset(self):
        return self._handle['entry/instrument/detector/data']

class ZebraHDF5Handler(HXNHandlerBase):
    HANDLER_NAME = "ZEBRA_HDF51_FLY_STREAM_V1"

    def __init__(self, resource_fn, *, frame_per_point):
        self._frame_per_point = frame_per_point
        super().__init__(resource_fn)

    def __call__(self, *, column, point_number):
        n_first = point_number * self._frame_per_point
        n_last = n_first + self._frame_per_point
        ds = self._handle[column]
        ds.id.refresh()
        return ds[n_first:n_last]

def register(db):
    db.reg.register_handler(SISHDF5Handler.HANDLER_NAME, SISHDF5Handler, overwrite=True)
    db.reg.register_handler(BulkMerlinStream.HANDLER_NAME, BulkMerlinStream,  overwrite=True)
    db.reg.register_handler(ZebraHDF5Handler.HANDLER_NAME, ZebraHDF5Handler, overwrite=True)




from databroker.v0 import Broker
from databroker.headersource.mongo import MDS
from databroker.assets.mongo import Registry
from databroker.headersource.core import doc_or_uid_to_uid
from databroker.assets.handlers import HandlerBase
from jsonschema import validate as js_validate
from collections import deque

import numpy as np
import pandas as pd
import warnings
import uuid
import certifi
import os
import six
import pymongo
import h5py
from datetime import datetime

import logging

# Load data from Eiger
db1_name = 'rs'
#db1_addr = 'mongodb://xf03id1-mdb01:27017,xf03id1-mdb02:27017,xf03id1-mdb03:27017'
db1_addr = 'mongodb://xf03id1-mdb02:27017,xf03id1-mdb03:27017'

_mds_config_db1 = {'host': db1_addr,
                   'port': 27017,
                   'database': 'datastore-2',
                   'timezone': 'US/Eastern'}

_fs_config_db1 = {'host': db1_addr,
                  'port': 27017,
                  'database': 'filestore-2'}

# bootstrap_servers = os.getenv("BLUESKY_KAFKA_BOOTSTRAP_SERVERS", None)
# kafka_password = os.getenv("BLUESKY_KAFKA_PASSWORD", None)
# try:
#     if bootstrap_servers is None or kafka_password is None:
#         raise Exception("BLUESKY_KAFKA_BOOTSTRAP_SERVERS or BLUESKY_KAFKA_PASSWORD environment variable not found.")
#     kafka_publisher = Publisher(
#             topic="hxn.bluesky.datum.documents",
#             bootstrap_servers=bootstrap_servers,
#             key=str(uuid.uuid4()),
#             producer_config={
#                     "acks": 1,
#                     "message.timeout.ms": 3000,
#                     "queue.buffering.max.kbytes": 10 * 1048576,
#                     "compression.codec": "snappy",
#                     "ssl.ca.location": certifi.where(),
#                     "security.protocol": "SASL_SSL",
#                     "sasl.mechanisms": "SCRAM-SHA-512",
#                     "sasl.username": "beamline",
#                     "sasl.password": kafka_password,
#                     },
#             flush_on_stop_doc=True,
#         ) if not os.environ.get('AZURE_TESTING') else None   # Disable on CI
# except:
#     print("Unable to setup kafka publisher, databroker will be readonly.")

try:
    f_benchmark = open("/nsls2/data/hxn/shared/config/bluesky/profile_collection/benchmark.out", "a+")
except:
    f_benchmark = None
datum_counts = {}

def sanitize_np(val):
    "Convert any numpy objects into built-in Python types."
    if isinstance(val, (np.generic, np.ndarray)):
        if np.isscalar(val):
            return val.item()
        return val.tolist()
    return val

def apply_to_dict_recursively(d, f):
    for key, val in d.items():
        if hasattr(val, 'items'):
            d[key] = apply_to_dict_recursively(val, f)
        d[key] = f(val)

def _write_to_file(col_name, method_name, t1, t2):
    if f_benchmark:
        f_benchmark.write(
            "{0}: {1}, t1: {2} t2:{3} time:{4} \n".format(
                col_name, method_name, t1, t2, (t2-t1),))
        f_benchmark.flush()

# class CompositeRegistry(Registry):
#     '''Composite registry.'''

#     def _register_resource(self, col, uid, spec, root, rpath, rkwargs,
#                               path_semantics):

#         run_start=None
#         ignore_duplicate_error=False
#         duplicate_exc=None

#         if root is None:
#             root = ''

#         resource_kwargs = dict(rkwargs)
#         if spec in self.known_spec:
#             js_validate(resource_kwargs, self.known_spec[spec]['resource'])

#         resource_object = dict(spec=str(spec),
#                                resource_path=str(rpath),
#                                root=str(root),
#                                resource_kwargs=resource_kwargs,
#                                path_semantics=path_semantics,
#                                uid=uid)

#         try:
#             col.insert_one(resource_object)
#         except Exception as duplicate_exc:
#             print(duplicate_exc)
#             if ignore_duplicate_error:
#                 warnings.warn("Ignoring attempt to insert Datum with duplicate "
#                           "datum_id, assuming that both ophyd and bluesky "
#                           "attempted to insert this document. Remove the "
#                           "Registry (`reg` parameter) from your ophyd "
#                           "instance to remove this warning.")
#             else:
#                 raise

#         resource_object['id'] = resource_object['uid']
#         resource_object.pop('_id', None)
#         ret = resource_object['uid']

#         return ret

#     def register_resource(self, spec, root, rpath, rkwargs,
#                               path_semantics='posix'):

#         uid = str(uuid.uuid4())
#         datum_counts[uid] = 0
#         method_name = "register_resource"
#         col = self._resource_col
#         ret = self._register_resource(col, uid, spec, root, rpath,
#                                       rkwargs, path_semantics=path_semantics)

#         return ret

#     def _insert_datum(self, col, resource, datum_id, datum_kwargs, known_spec,
#                      resource_col, ignore_duplicate_error=False,
#                      duplicate_exc=None):
#         if ignore_duplicate_error:
#             assert duplicate_exc is not None
#         if duplicate_exc is None:
#             class _PrivateException(Exception):
#                 pass
#             duplicate_exc = _PrivateException
#         try:
#             resource['spec']
#             spec = resource['spec']

#             if spec in known_spec:
#                 js_validate(datum_kwargs, known_spec[spec]['datum'])
#         except (AttributeError, TypeError):
#             pass
#         resource_uid = self._doc_or_uid_to_uid(resource)
#         if type(datum_kwargs) == str and '/' in datum_kwargs:
#             datum_kwargs = {'point_number': datum_kwargs.split('/')[-1]}

#         datum = dict(resource=resource_uid,
#                      datum_id=str(datum_id),
#                      datum_kwargs=dict(datum_kwargs))
#         apply_to_dict_recursively(datum, sanitize_np)
#         # We are transitioning from ophyd objects inserting directly into a
#         # Registry to ophyd objects passing documents to the RunEngine which in
#         # turn inserts them into a Registry. During the transition period, we allow
#         # an ophyd object to attempt BOTH so that configuration files are
#         # compatible with both the new model and the old model. Thus, we need to
#         # ignore the second attempt to insert.
#         try:
#             kafka_publisher('datum', datum)
#             #col.insert_one(datum)
#         except duplicate_exc:
#             if ignore_duplicate_error:
#                 warnings.warn("Ignoring attempt to insert Resource with duplicate "
#                               "uid, assuming that both ophyd and bluesky "
#                               "attempted to insert this document. Remove the "
#                               "Registry (`reg` parameter) from your ophyd "
#                               "instance to remove this warning.")
#             else:
#                 raise
#         # do not leak mongo objectID
#         datum.pop('_id', None)

#         return datum


#     def register_datum(self, resource_uid, datum_kwargs, validate=False):

#         if validate:
#             raise RuntimeError('validate not implemented yet')

#         res_uid = resource_uid
#         datum_count = datum_counts[res_uid]

#         datum_uid = res_uid + '/' + str(datum_count)
#         datum_counts[res_uid] = datum_count + 1

#         col = self._datum_col
#         datum = self._insert_datum(col, resource_uid, datum_uid, datum_kwargs, {}, None)
#         ret = datum['datum_id']

#         return ret

#     def _doc_or_uid_to_uid(self, doc_or_uid):

#         if not isinstance(doc_or_uid, six.string_types):
#             try:
#                 doc_or_uid = doc_or_uid['uid']
#             except TypeError:
#                 pass

#         return doc_or_uid

#     def _bulk_insert_datum(self, col, resource, datum_ids,
#                            datum_kwarg_list):

#         resource_id = self._doc_or_uid_to_uid(resource)

#         to_write = []

#         d_uids = deque()

#         for d_id, d_kwargs in zip(datum_ids, datum_kwarg_list):
#             dm = dict(resource=resource_id,
#                       datum_id=str(d_id),
#                       datum_kwargs=dict(d_kwargs))
#             apply_to_dict_recursively(dm, sanitize_np)
#             to_write.append(pymongo.InsertOne(dm))
#             d_uids.append(dm['datum_id'])

#         col.bulk_write(to_write, ordered=False)

#         return d_uids

#     def bulk_register_datum_table(self, resource_uid, dkwargs_table, validate=False):

#         res_uid = resource_uid['uid']
#         datum_count = datum_counts[res_uid]

#         if validate:
#             raise RuntimeError('validate not implemented yet')

#         d_ids = [res_uid + '/' + str(datum_count+j) for j in range(len(dkwargs_table))]
#         datum_counts[res_uid] = datum_count + len(dkwargs_table)

#         dkwargs_table = pd.DataFrame(dkwargs_table)
#         datum_kwarg_list = [ dict(r) for _, r in dkwargs_table.iterrows()]

#         method_name = "bulk_register_datum_table"

#         self._bulk_insert_datum(self._datum_col, resource_uid, d_ids, datum_kwarg_list)
#         return d_ids


# mds_db1 = MDS(_mds_config_db1, auth=False)
# db1 = Broker(mds_db1, CompositeRegistry(_fs_config_db1))

# wrapper for two databases
class CompositeBroker(Broker):
    """wrapper for two databases"""
    # databroker.headersource.MDSROTemplate
    def _bulk_insert_events(self, event_col, descriptor, events, validate, ts):

        descriptor_uid = doc_or_uid_to_uid(descriptor)

        to_write = []
        for ev in events:
            data = dict(ev['data'])

            # Replace any filled data with the datum_id stashed in 'filled'.
            for k, v in six.iteritems(ev.get('filled', {})):
                if v:
                    data[k] = v
            # Convert any numpy types to native Python types.
            apply_to_dict_recursively(data, sanitize_np)
            timestamps = dict(ev['timestamps'])
            apply_to_dict_recursively(timestamps, sanitize_np)

            # check keys, this could be expensive
            if validate:
                if data.keys() != timestamps.keys():
                    raise ValueError(
                        BAD_KEYS_FMT.format(data.keys(),
                                            timestamps.keys()))
            ev_uid = ts + '-' + ev['uid']

            ev_out = dict(descriptor=descriptor_uid, uid=ev_uid,
                          data=data, timestamps=timestamps,
                          time=ev['time'],
                          seq_num=ev['seq_num'])

            to_write.append(pymongo.InsertOne(ev_out))

        event_col.bulk_write(to_write, ordered=True)

    # databroker.headersource.MDSROTemplate
    # databroker.headersource.MDSRO(MDSROTemplate)
    def _insert(self, name, doc, event_col, ts):
        for desc_uid, events in doc.items():
            # If events is empty, mongo chokes.
            if not events:
                continue
            self._bulk_insert_events(event_col,
                                     descriptor=desc_uid,
                                     events=events,
                                     validate=False, ts=ts)


    def insert(self, name, doc):

        if name == "start":
            if f_benchmark:
                f_benchmark.write("\n scan_id: {} \n".format(doc['scan_id']))
                f_benchmark.flush()
            datum_counts = {}

        ts =  str(datetime.now().timestamp())

        if name in {'bulk_events'}:
            ret2 = self._insert(name, doc, self.mds._event_col, ts)
        elif name == 'event_page':
            import event_model
            for ev_doc in event_model.unpack_event_page(doc):
                super().insert('event', ev_doc)
            ret2 = None
        else:
            ret2 = super().insert(name, doc)
        return ret2

logger = logging.getLogger(__name__)

FMT_ROI_KEY = 'entry/instrument/detector/NDAttributes/CHAN{}ROI{}'
from databroker.assets.handlers import HandlerBase, ImageStack


class HDF5DatasetSliceHandler(HandlerBase):
    """
    Handler for data stored in one Dataset of an HDF5 file.
    Parameters
    ----------
    filename : string
        path to HDF5 file
    key : string
        key of the single HDF5 Dataset used by this Handler
    frame_per_point : integer, optional
        number of frames to return as one datum, default 1
    swmr : bool, optional
        Open the hdf5 file in SWMR read mode. Only used when mode = 'r'.
        Default is False.
    """
    def __init__(self, filename, key, frame_per_point=1):
        self._fpp = frame_per_point
        self._filename = filename
        self._key = key
        self._file = None
        self._dataset = None
        self._data_objects = {}
        self.open()

    def get_file_list(self, datum_kwarg_gen):
        return [self._filename]

    def __call__(self, point_number):
        # Don't read out the dataset until it is requested for the first time.
        if not self._dataset:
            self._dataset = self._file[self._key]

        if point_number not in self._data_objects:
            start = point_number * self._fpp
            stop = (point_number + 1) * self._fpp
            self._data_objects[point_number] = ImageStack(self._dataset,
                                                          start, stop)
        return self._data_objects[point_number]

    def open(self):
        import h5py
        if self._file:
            return

        self._file = h5py.File(self._filename, 'r')

    def close(self):
        super(HDF5DatasetSliceHandler, self).close()
        self._file.close()
        self._file = None

def HXN_compose_db(reg=Registry):
    mds_db1 = MDS(_mds_config_db1, auth=False)
    db = CompositeBroker(mds_db1, reg(_fs_config_db1))
    db.name = 'hxn'

    try:
        from .handlers import register
        register(db)
    except:
        pass

    return db

# mds_db1 = MDS(_mds_config_db1, auth=False)
# db1 = Broker(mds_db1, CompositeRegistry(_fs_config_db1))
# db = CompositeBroker(mds_db1, CompositeRegistry(_fs_config_db1))
# db.name = 'hxn'

db = HXN_compose_db()

def get_path(scan_id,key_name,db=db):
    return db.reg.get_spec_handler(db[scan_id].table(fields=[key_name])[key_name][1].split('/')[0]).filename

# try:
#     from hxntools.handlers import register
#     register(db)
# except:
#     pass



motor_table = {'zpssx':('inenc2_val',-1e-4*0.9542,'/INENC2.VAL.Value'), #-9.7e-5), # Additional correction ratio here should match with changes to the CapSensorVtoUM scaling factor in Power PMAC config files to ensure scan input and readback positions match.
            'zpssy':('inenc3_val',-1.03297e-4,'/INENC3.VAL.Value'), #-1.006e-4),
            'zpssz':('inenc4_val',1e-4,'/INENC4.VAL.Value'), #1.04e-4),
            'dssx':('inenc2_val',-1e-4,'/INENC2.VAL.Value'),
            'dssy':('inenc3_val',1e-4,'/INENC3.VAL.Value'),
            'dssz':('inenc4_val',1e-4*1.047,'/INENC4.VAL.Value'), # See above comment for correction ratio.

            # ss* stages' position readout cannot be collected by PandABox encoder inputs, make the position 1 here and calculate with softwaer instead
            'ssx':('inenc2_val',1e-4,'/INENC2.VAL.Value'),
            'ssy':('inenc3_val',1e-4,'/INENC3.VAL.Value'),
            'ssz':('inenc4_val',1e-4,'/INENC4.VAL.Value'), # See above comment for correction ratio.

            # pt_tomo_ssx conversion ratio prior to Nov. 2025 was set to -1*current value because the image recorded from eiger1_mobile detector was not mirrored
            'pt_tomo_ssx':('inenc2_val',-1e-4,'/INENC2.VAL.Value'),
            'pt_tomo_ssy':('inenc4_val',-1e-4,'/INENC4.VAL.Value')

            # Stage encoder
            #'pt_tomo_ssy':('inenc4_val',6e-5)

            # Scanning MLL setup
            #'pt_tomo_ssx':('inenc1_val',-1e-4, '/INENC1.VAL.Value'),
            #'pt_tomo_ssy':('inenc2_val',-1e-4, '/INENC2.VAL.Value')
            }



from __future__ import print_function

import logging

# imort XRF_DATA_KEY for back-compat
from databroker.assets.handlers import (Xspress3HDF5Handler,
                                        XS3_XRF_DATA_KEY as XRF_DATA_KEY)


logger = logging.getLogger(__name__)

FMT_ROI_KEY = 'entry/instrument/detector/NDAttributes/CHAN{}ROI{}'

class XSP3HDF5Handler(Xspress3HDF5Handler):
    def __init__(self,filename,key=XRF_DATA_KEY):
        if filename.startswith('/data'):
            filename = '/nsls2/data/hxn/legacy' + filename[5:]
        super().__init__(filename,key)

def register(db):
    #db.reg.register_handler(Xspress3HDF5Handler.HANDLER_NAME,
    #                        Xspress3HDF5Handler)
    db.reg.register_handler(XSP3HDF5Handler.HANDLER_NAME,
                            XSP3HDF5Handler, overwrite=True)



from .hxn_handler import HXNHandlerBase
from databroker.assets.handlers import HandlerBase
import h5py


class BulkMerlin(HXNHandlerBase):
    HANDLER_NAME = 'MERLIN_HDF5_BULK'

    def __call__(self, **kwargs):
        ds = self._handle['entry/instrument/detector/data']
        ds.id.refresh()
        return ds[:,:,:]

    def dataset(self):
        return self._handle['entry/instrument/detector/data']

class BulkXSP(HXNHandlerBase):
    HANDLER_NAME = 'XSP3_BULK'

    def __call__(self, frame = None, channel = None):
        ds = self._handle['entry/instrument/detector/data']
        ds.id.refresh()
        if channel is None:
            return ds[:,:,:]
        else:
            return ds[:,channel-1,:].squeeze()

class ROIHDF5Handler(HXNHandlerBase):
    HANDLER_NAME = "ROI_HDF5_FLY"

    def __call__(self, *, det_elem):
        ds = self._handle[det_elem]
        ds.id.refresh()
        return ds[:]

    def close(self):
        self._handle.close()
        self._handle = None
        super().close()

class PandAHandlerHDF5(HXNHandlerBase):
    """The handler to read HDF5 files produced by PandABox."""
    HANDLER_NAME = "PANDA"

    specs = {"PANDA"}

    def __call__(self, field):
        ds = self._handle[f"/{field}"]
        return ds[:]

class DexelaHandlerHDF5(HXNHandlerBase):
    """The handler to read HDF5 files produced by PandABox."""
    HANDLER_NAME = "DEX_HDF5"

    def __call__(self, frame):
        ds = self._handle['entry/instrument/detector/data']
        ds.id.refresh()
        return ds[frame,:,:]



def register(db):
    db.reg.register_handler(BulkMerlin.HANDLER_NAME,
                            BulkMerlin, overwrite=True)
    db.reg.register_handler(BulkXSP.HANDLER_NAME,
                            BulkXSP, overwrite=True)
    db.reg.register_handler(ROIHDF5Handler.HANDLER_NAME,
                            ROIHDF5Handler, overwrite=True)
    db.reg.register_handler(PandAHandlerHDF5.HANDLER_NAME,
                            PandAHandlerHDF5, overwrite=True)
    db.reg.register_handler(DexelaHandlerHDF5.HANDLER_NAME,
                            DexelaHandlerHDF5, overwrite=True)
